#include "mbed.h"
#include "system.h"
#include "display.h"
#include "camera.h"
#include "phone.h"
#include "canvas.h"

#define LEPTON_25_MHZ 0,9
#define LEPTON_CS_B  0,6
#define LEPTON_RESET_B  2,12
#define LEPTON_PWDN_B  2,13
#define LEPTON_MISO 0,8
#define LEPTON_SCK 0,7
#define LEPTON_SCL 0,28
#define LEPTON_SDA 0,27
#define LEPTON_I2C_ADDR (0x2a)


#define LEPTON_LINELENGTH 164

unsigned char camera_linebuffer[LEPTON_LINELENGTH+2]; // the two last, extra bytes are zeros for CRC
unsigned short camera_image[80*60];
unsigned char camera_index_image[80*60],camera_index_snap[80*60];
unsigned short camera_palette[256];
unsigned char camera_palettepoints[9][3]={{0,0,0},{0,0,255},{128,0,255},{255,0,255},{255,0,128},{255,0,0},{255,128,0},{255,255,0},{255,255,255}};
long camera_snapshot=0;

volatile long camera_ready=0;
volatile long camera_synccount=0;
volatile long camera_line;
volatile long camera_newimage=0;
volatile long camera_tick=0;
volatile long camera_time=0;
long camera_crc_errorcount=0;
long camera_sum=0,camera_lastsum=0;
long camera_min,camera_max,camera_scale;

void camera_readline(void);
void camera_index(void);
unsigned char camera_writebyte(unsigned char data);
void camera_resync(void);
unsigned short camera_checksum(void);
extern "C" void camera_interrupt(void);

int camera_i2c_write(uint16_t address, int length, uint16_t *data);
int camera_i2c_read(uint16_t address, int length, uint16_t *data);
int camera_i2c_write_byte(unsigned char data);
unsigned char camera_i2c_read_byte(int last);
int camera_i2c_status;



// The Lepton pins are set to pull_down mode to make sure there is no problem
// during powerdown of the unit
// The Lepton I2C command bus is connected, but not used at this time.
// The SDA and SCL lines have external pull-ups (as they should)
void camera_init(void)
   {
   long i,j;
   // the main clock is 96 MHz, which divided by 4 is 24 MHz, which will do for the Lepton 25 MHz clock
   system_configurepin(LEPTON_25_MHZ, 3, 0, 0, 0);  // p76, P0.9, 3 (MAT2.3)
   system_setbitpair(&LPC_SC->PCLKSEL1,12,1);       // 12 = PCLK_TIMER2 (1 = CCLK)
   LPC_TIM2->MCR=0x00000400;                        // clear on match 3
   LPC_TIM2->MR3=0x00000001;                        // match 3 every 2nd cycle
   LPC_TIM2->EMR=0x00000c00;                        // toggle match 3 pin
   LPC_TIM2->TCR=0x00000002;                        // reset timer
   LPC_TIM2->TCR=0x00000001;                        // enable timer
   
   // the Lepton video output is through SSP1
   system_setbitpair(&LPC_SC->PCLKSEL0,20,1);       // 20 = PCLK_SSP1 (1 = CCLK)
   LPC_SSP1->CPSR=6;                                // try 4 MHz clock (max for Lepton is 20 MHz, min 2 MHz). The documentation claims this to be the actual divisor, not divisor-1
   LPC_SSP1->CR0=0x000000c7;                        // 8-bit transfers, SPOL = 1, SPHA = 1
   LPC_SSP1->CR1=0x00000002;                        // enable
   system_configurepin(LEPTON_MISO, 2, 0, 0, 0); // p77, P0.8, 2 (MISO1)
   system_configurepin(LEPTON_SCK, 2, 0, 0, 0);     // p78, P0.7, 2 (SCK1)

   // the Lepton control is done through I2C0 (Lepton does work without control, but we want to calibrate it)
   system_setbitpair(&LPC_SC->PCLKSEL0,14,1);       // 14 = PCLK_I2C0 (1 = CCLK)
   LPC_I2C0->I2CONSET=0x00000040;                   // enable
   LPC_I2C0->I2CONCLR=0x00000038;                   // clear STA, STO and SI
   LPC_I2C0->I2SCLL=120;                            // low clock phase 1.25 us
   LPC_I2C0->I2SCLH=120;                            // high clock phase 1.25 us
   
   system_configurepin(LEPTON_SDA, 1, 0, 0, 0);     // p25, P0.27, 1 (SDA0)
   system_configurepin(LEPTON_SCL, 1, 0, 0, 0);     // p24, P0.28, 1 (SCL0)

   
   // LEPTON_CS_B
   system_configurepin(LEPTON_CS_B, 0, 3, 1, 1);    // p79, P0.6, 0 (GPIO), pull-down, output, 1
   
   // LEPTON_RESET_B
   system_configurepin(LEPTON_RESET_B, 0, 3, 1, 1); // p51, P2.12, 0 (GPIO), pull-down, output, 1
   
   // LEPTON_PWDN_B
   system_configurepin(LEPTON_PWDN_B, 0, 3, 1, 1);  // p50, P2.13, 0 (GPIO), pull-down, output, 1
   

   
   // boot camera
   SETIO(LEPTON_PWDN_B);
   CLRIO(LEPTON_RESET_B);
   SETIO(LEPTON_CS_B);
   wait_ms(100); // 1 should be enough
   SETIO(LEPTON_RESET_B);
   
   // Lepton is now initially reset and resynced  
   camera_linebuffer[LEPTON_LINELENGTH]=0;
   camera_linebuffer[LEPTON_LINELENGTH+1]=0; // prepare for CRC, which pumps these two zeros through the shift register
   
   // launch camera interrupt service - it will not do anything yet as camera_ready==0
   NVIC_SetVector(TIMER1_IRQn, (uint32_t)&camera_interrupt);
   NVIC_EnableIRQ(TIMER1_IRQn);
   system_setbitpair(&LPC_SC->PCLKSEL0,4,1);  // 4 = PCLK_TIMER1 (1 = CCLK)
   LPC_TIM1->MCR=0x00000003;                  // clear and interrupt on match 0
   LPC_TIM1->MR0=(96000/2)-1;                 // match 0 at 2 kHz
   LPC_TIM1->TCR=0x00000002;                  // reset timer
   LPC_TIM1->TCR=0x00000001;                  // enable timer (and the interrupt)

   camera_ready=2000;

   // prepare palette
   for(i=0;i<256;i++)
      {
      unsigned char r,g,b,r1,g1,b1,r2,g2,b2;
      r1=camera_palettepoints[i>>5][0];
      g1=camera_palettepoints[i>>5][1];
      b1=camera_palettepoints[i>>5][2];
      r2=camera_palettepoints[1+(i>>5)][0];
      g2=camera_palettepoints[1+(i>>5)][1];
      b2=camera_palettepoints[1+(i>>5)][2];
      r=(r1*(((31-i)&31)+1)+r2*(i&31))>>5;
      g=(g1*(((31-i)&31)+1)+g2*(i&31))>>5;
      b=(b1*(((31-i)&31)+1)+b2*(i&31))>>5;
      camera_palette[i]=(((r>>3)<<11)|((g>>2)<<5)|((b>>3)));
      }
   for(i=0;i<80*60;i++) camera_index_snap[i]=0;
   camera_calibrate();
   }


uint16_t camera_calibrationcommand[]={0x0004,0x0242};
void camera_calibrate(void)
   {
   camera_i2c_write(2,camera_calibrationcommand);
   }
       
       
       
int camera_i2c_write(int length, uint16_t *data)
   {
   int i;
   LPC_I2C0->I2CONCLR=0x00000038;   // clear STA, STO and SI
   LPC_I2C0->I2CONSET=0x00000020;   // start
   wait_us(10);                     // no other masters, this cannot take longer
   camera_i2c_status=LPC_I2C0->I2STAT;
   if (((LPC_I2C0->I2STAT)&0xf8)!=0x08) return(1);   // start ok?
   LPC_I2C0->I2DAT=(LEPTON_I2C_ADDR<<1);    // write address + 'write'-bit
   LPC_I2C0->I2CONCLR=0x00000038;   // clear STA, STO and SI
   wait_us(30);
   if (((LPC_I2C0->I2STAT)&0xf8)!=0x18) return(2);   // address acknowledged ok?
   for(i=0;i<length;i++)
      {
      if (camera_i2c_write_byte((*data>>8)&0xff)) return(3);
      if (camera_i2c_write_byte((*data)&0xff)) return(3);
      data++;
      }
   LPC_I2C0->I2CONCLR=0x00000038;   // clear STA, STO and SI
   LPC_I2C0->I2CONSET=0x00000010;   // stop
   return(0);
   }
   
   
   
int camera_i2c_read(int length, uint16_t *data)
   {
   unsigned char hb;
   int i;
   LPC_I2C0->I2CONCLR=0x00000038;   // clear STA, STO and SI
   LPC_I2C0->I2CONSET=0x00000020;   // start
   wait_us(10);                     // no other masters, this cannot take longer
   if (((LPC_I2C0->I2STAT)&0xf8)!=0x08) return(1);   // start ok?
   LPC_I2C0->I2DAT=(LEPTON_I2C_ADDR<<1)|1;    // write address + 'read'-bit
   LPC_I2C0->I2CONCLR=0x00000038;   // clear STA, STO and SI
   wait_us(30);
   if (((LPC_I2C0->I2STAT)&0xf8)!=0x40) return(2);   // address acknowledged ok?
   for(i=0;i<length;i++)
      {
      hb=camera_i2c_read_byte(0);
      *data=(hb<<8)|(camera_i2c_read_byte(i==length-1)&0xff);
      data++;
      }
   LPC_I2C0->I2CONCLR=0x00000038;   // clear STA, STO and SI
   LPC_I2C0->I2CONSET=0x00000010;   // stop
   return(0);
   }
   
   
   
int camera_i2c_write_byte(unsigned char data)
   {
   LPC_I2C0->I2DAT=data;
   LPC_I2C0->I2CONCLR=0x00000038;   // clear STA, STO and SI
   wait_us(30);
   if (((LPC_I2C0->I2STAT)&0xf8)!=0x28) return(1);   // data acknowledged ok?
   return(0);
   }
   
   
   
unsigned char camera_i2c_read_byte(int last)
   {
   unsigned char data;
   if (last)
      {
      LPC_I2C0->I2CONCLR=0x0000003c;   // clear STA, STO, SI and AA
      }
   else
      {
      LPC_I2C0->I2CONCLR=0x00000038;   // clear STA, STO and SI
      LPC_I2C0->I2CONSET=0x00000004;   // prepare to NACK the last data byte
      }
   wait_us(3000);
   //if (((LPC_I2C0->I2STAT)&0xf8)!=0x50) return(1); 
   data=LPC_I2C0->I2DAT;
   return(data);
   }
   
            
            
// camera IRQ - called at 2 kHz to keep Lepton happy
extern "C" void camera_interrupt(void)
   {
   camera_tick=1;
   LPC_TIM1->IR=LPC_TIM1->IR; // clear bit(s) in interrupt register
   if (!camera_ready) return; // not started yet
   if (camera_synccount)      // resync in progress
      {
      camera_synccount--;
      return;
      }
   if (camera_ready>1) // The Lepton needs this early reboot after 1000 lines. This is quite ugly, but what should you do? 
      {
      camera_ready--;
      if (camera_ready==1)
         {
         camera_synccount=400; // 200 ms delay to cause Lepton time-out
         return;
         }
      }
   camera_readline();
   camera_time=LPC_TIM1->TC;
   }
   
   
   
void camera_readline(void)
   {
   int i,value;
   CLRIO(LEPTON_CS_B);
   while (LPC_SSP1->SR&0x00000010); // wait for the controller to finish the previous task (there should be none!)
   while (LPC_SSP1->SR&0x00000004) value=LPC_SSP1->DR; // flush the receive FIFO (should be empty!)
   for(i=0;i<LEPTON_LINELENGTH;i++)
      { // basically the camera_writebyte(0) inlined here
      LPC_SSP1->DR=0; // send the dummy byte
      while (LPC_SSP1->SR&0x00000010);
      camera_linebuffer[i]=LPC_SSP1->DR;
      }
   SETIO(LEPTON_CS_B);
   
   // only check crc for actual lines (it is invalid for the discard-lines, anyway)
   // it seems that the lepton generates one crc error per new (true) image
   // (and always for line 0x2f or 0x30 with this data rate)
   if ((camera_linebuffer[0]&0x0f)!=0x0f) // genuine data packet
      {
      unsigned char a,b;
      unsigned short cs;
      a=camera_linebuffer[2];
      b=camera_linebuffer[3];
      cs=camera_checksum();
      if ((a==((cs>>8)&0xff))&&(b==(cs&0xff))) // checksum is correct
         {
         camera_crc_errorcount=0;
         if (((camera_linebuffer[0]&0x0f)==0)&&(camera_linebuffer[1]<60))
            {
            camera_line=camera_linebuffer[1];
            if (camera_line==0) camera_sum=0;
            for(i=0;i<80;i++)
               {
               unsigned short pixel;
               pixel=(camera_linebuffer[4+2*i]<<8)+camera_linebuffer[4+2*i+1];
               camera_image[80*camera_line+i]=pixel;
               camera_sum+=pixel;
               }
            if ((camera_line==59)&&(camera_sum!=camera_lastsum))
               {
               camera_newimage=1;
               camera_lastsum=camera_sum;
               }
            }
         }
      else // checksum error
         {
         camera_crc_errorcount++;
         if (camera_crc_errorcount>2) // one error is normal
            {
            camera_synccount=400; // 200 ms delay
            camera_crc_errorcount=0;
            }
         }
      }
   }



unsigned char camera_writebyte(unsigned char data) // the data is actually ignored by the Lepton connection
   {
   unsigned char value;
   while (LPC_SSP1->SR&0x00000010); // wait for the controller to finish the previous task (there should be none!)
   while (LPC_SSP1->SR&0x00000004) value=LPC_SSP1->DR; // flush the receive FIFO (should be empty!)
   LPC_SSP1->DR=data; // send the data
   while (LPC_SSP1->SR&0x00000010);
   //while (LPC_SSP1->SR&0x00000004==0); // wait for the response to show up in the FIFO - for some reason this does not work
   return(LPC_SSP1->DR);
   }



// TODO: optimize
unsigned short camera_checksum(void)
   {
   unsigned short crc=0x0000;
   int i,j,flag;

   camera_linebuffer[0]&=0x0f;
   camera_linebuffer[2]=0;
   camera_linebuffer[3]=0;
   for(i=0;i<LEPTON_LINELENGTH+2;i++)
      {
      for(j=7;j>=0;j--)
         {
         flag=(crc&0x8000) ? 1 : 0;
         crc<<=1;
         if (camera_linebuffer[i]&(1<<j)) crc++;
         if (flag) crc^=0x1021;
         }
       }
   return(crc);
   }



void camera_index(void)
   {
   int i;
   unsigned short pixel;
   camera_min=camera_max=camera_image[0];
   for (i=1;i<80*60;i++)
      {
      pixel=camera_image[i];
      if (camera_min>pixel) camera_min=pixel;
      if (camera_max<pixel) camera_max=pixel;
      }
   if (camera_max==camera_min) camera_max++; // avoid division by zero
   camera_scale=0x00ffffff/(camera_max-camera_min);
   for (i=0;i<80*60;i++) camera_index_image[i]=((camera_image[i]-camera_min)*camera_scale)>>16;    
   }
    
    
void camera_snap(void)
   {
   int i;
   for(i=0;i<80*60;i++) camera_index_snap[i]=camera_index_image[i];
   camera_snapshot=1;
   }
       