#include "mbed.h"
#include "fun2016.h"
#include "camera.h"
#include "system.h"
#include "util.h"
#include "gsm.h"


enum GSM_STATE {GSM_MODE_BOOT,GSM_MODE_READY,GSM_MODE_OFF};

#define INBUFFERSIZE 32
MODSERIAL gsm(P2_0,P2_1);    // SIM900D

volatile long gsm_ringing=0,gsm_hangup=0,gsm_online=0;
int gsm_pinready=0,gsm_waitingforpin=0,gsm_rebootcount=0;
char gsm_pin[5];
long gsm_time,gsm_mode,gsm_wait,gsm_submode; // submode was originally for booting sequence, but others benefit from it, too
long gsm_insisting,gsm_insist_starttime,gsm_insist_timeout,gsm_insist_repeatcount,gsm_insist_successmode,gsm_insist_successsubmode,gsm_insist_failmode,gsm_insist_failsubmode;
volatile long gsm_insist_success;
char gsm_insist_command[64],gsm_insist_hope[32];

char gsm_inbuffer[INBUFFERSIZE];
long gsm_inbuffercount=0;

void gsm_insist(char *command, char *hope, int timeout_us, int repeatcount, int successmode, int successsubmode, int failmode, int failsubmode);
void gsm_receive(MODSERIAL_IRQ_INFO *q);
int gsm_try(char *command, char *hope, int timeout_us);
void gsm_writetif(void);
void gsm_writegif(void);
void gsm_wait_putc(char data);

void gsm_init(void)
   {
   // the SIM900D is connected to the UART1 - which is currently handled by MBED

   system_configurepin(GSM_PWRKEY, 0, 0, 1, 1);    // p73, P2.2, 0 (GPIO), pull-up, output, 1
   
   system_configurepin(GSM_STATUS, 0, 0, 0, 0);    // p70, P2.3, 0 (GPIO), pull-up, input

   system_configurepin(GSM_NETLIGHT, 0, 0, 0, 0);    // p69, P2.4, 0 (GPIO), pull-up, input

   system_configurepin(GSM_RI, 0, 0, 0, 0);    // p67, P2.6, 0 (GPIO), pull-up, input

   system_configurepin(GSM_RTS, 0, 0, 1, 0);    // p9, P0.23, 0 (GPIO), pull-up, output, 0 (1 = wait)

   system_configurepin(GSM_CTS, 0, 0, 0, 0);    // p8, P0.24, 0 (GPIO), pull-up, input

   gsm_time=LPC_TIM3->TC;
   gsm_mode=0;
   gsm_submode=0;
   gsm_wait=0;
   gsm_rebootcount=0;
   
   // attach an interrupt routine to sniff the gsm responses
   gsm.attach(&gsm_receive, MODSERIAL::RxIrq);
   }



// TODO: ignore binary data
void gsm_receive(MODSERIAL_IRQ_INFO *q)
   {
   while (gsm.readable())
      {
   char c,a,b;
   if (gsm_inbuffercount>INBUFFERSIZE-2) gsm_inbuffercount=0; // overflow, something is amiss, so could as well start from the beginning
   c=gsm.getc();
   // echo to debug - cannot use prinf
   //debug.putc('>');
   util_makeprintable(c,&a,&b);
   debug.putc(a);
   if (b) debug.putc(b);
   if (b=='n') debug.putc('\n'); // real newline after the transcribed newline
   //debug.putc('<');
   gsm_inbuffer[gsm_inbuffercount++]=c;
   if (c=='\n')
      {
      gsm_inbuffer[gsm_inbuffercount]=0; // make it a string
      gsm_inbuffercount=0;
      if (gsm_insisting)
         {
         if (!strcmp(gsm_inbuffer,gsm_insist_hope))
            {
            gsm_insist_success=1;
            }
         }
      if (!strcmp(gsm_inbuffer,"RING\r\n"))
         {
         gsm_ringing=1;
         gsm_hangup=0;
         }
      if ((!strcmp(gsm_inbuffer,"NO CARRIER\r\n"))||(!strcmp(gsm_inbuffer,"BUSY\r\n")))
         {
         gsm_hangup=1;
         gsm_ringing=0;
         }
      if (!strcmp(gsm_inbuffer,"Call Ready\r\n"))
         {
         gsm_online=1;
         gsm_hangup=0;
         gsm_ringing=0;
         debug.printf("GSM now online!\n");
         }
      }
   }
   }



void gsm_update()
   {
   long gsm_step;
   gsm_step=LPC_TIM3->TC-gsm_time;
   gsm_time+=gsm_step;
   gsm_wait-=gsm_step;
   // check for waiting first
   if (gsm_wait>0) return;
   // check for insisting
   if (gsm_insisting)
      {
      if (gsm_insist_success)
         {
         gsm_insisting=0;
         gsm_mode=gsm_insist_successmode;
         gsm_submode=gsm_insist_successsubmode;
         }
      else
         {
         if (gsm_insist_repeatcount)
            {
            debug.printf("Keep insisting:%s",gsm_insist_command);
            gsm.printf("%s",gsm_insist_command);
            gsm_wait=gsm_insist_timeout;
            gsm_insist_repeatcount--;
            return;
            }
         else // insist failed
            {
            gsm_insisting=0;
            gsm_mode=gsm_insist_failmode;
            gsm_submode=gsm_insist_failsubmode;
            }
         }
      }
   // finally run the state machine
   switch(gsm_mode)
      {
      case GSM_MODE_BOOT: // booting
         switch(gsm_submode)
            {
            case 0: // beginning boot sequence, press powerkey
               CLRIO(GSM_PWRKEY);
               gsm_wait=1500000;
               gsm_submode=1;
               gsm_pinready=0;
               debug.printf("Pressing POWERKEY\n");
               return;
               break;
            case 1: // release powerkey - wait for GSM_STATUS to resolve
               SETIO(GSM_PWRKEY);
               gsm_wait=2500000;
               gsm_submode=2;
               debug.printf("Releasing POWERKEY\n");
               return;
               break;              
            case 2: // check GSM_STATUS to resolve whether the module turned on or off (it is entirely OK for it to turn off to recover from a serious error)
               if (GETIO(GSM_STATUS))
                  {
                  gsm_insist("AT\r","OK\r\n",1000000,5,GSM_MODE_BOOT,3,GSM_MODE_BOOT,0);                 
                  return;
                  }
               else
                  {
                  gsm_submode=0; // reboot 
                  gsm_rebootcount++;
                  if (gsm_rebootcount>2) gsm_mode=GSM_MODE_OFF; // most likely we are running solely on USB power, so the module cannot start - stop trying
                  return;
                  }                 
               break;
            case 3: // set fixed baud rate to enable status messages
               gsm_insist("AT+IPR=9600\r","OK\r\n",1000000,5,GSM_MODE_BOOT,4,GSM_MODE_BOOT,0);                 
               return;
            case 4: // ask the PIN code
               gsm_waitingforpin=1;
               if (!gsm_pinready) return;     // stay here until we have the pin
               gsm_waitingforpin=0;
               sprintf(gsm_insist_command,"AT+CPIN=%s\r",gsm_pin);
               gsm_insist(gsm_insist_command,"OK\r\n",1000000,5,GSM_MODE_BOOT,5,GSM_MODE_BOOT,0);
               //gsm_insist("AT+CPIN=2915\r","OK\r\n",1000000,5,GSM_MODE_BOOT,5,GSM_MODE_BOOT,0);                 
               return;
            case 5: // enable CTS/RTS flow control (DTR part)
               gsm_insist("AT&D1\r","OK\r\n",1000000,5,GSM_MODE_BOOT,6,GSM_MODE_BOOT,0);                 
               return;
            case 6: // enable CTS/RTS flow control
               gsm_insist("AT+IFC=2,2\r","OK\r\n",1000000,5,GSM_MODE_BOOT,7,GSM_MODE_BOOT,0);                 
               return;             
            case 7: // set sound level to full
               gsm_insist("AT+CLVL=100\r","OK\r\n",1000000,5,GSM_MODE_BOOT,8,GSM_MODE_BOOT,0);                 
               return;
            case 8: // set ring tone level to full
               gsm_insist("AT+CRSL=3\r","OK\r\n",1000000,5,GSM_MODE_READY,0,GSM_MODE_BOOT,0);                 
               return;
            default: // should never happen!
               debug.printf("Unknown GSM mode %d,%d - rebooting\n");
               gsm_mode=0;
               gsm_submode=0;
               return;
               break;
            }
         break;
      case GSM_MODE_READY: // ready for call
         return;
         break;
      case GSM_MODE_OFF: // just do nothing
         return;
         break;
      default: // should never happen!
         debug.printf("Unknown GSM mode %d,%d - rebooting\n");
         gsm_mode=0;
         gsm_submode=0;
         return;
         break;
      }
   }
   
// max 31 chars for command and hope!
// make a command and hope for a nice answer. Action happens in the gsm_update(). Any input from the module is checked against the hope
// as the characters arrive
void gsm_insist(char *command, char *hope, int timeout_us, int repeatcount, int successmode, int successsubmode, int failmode, int failsubmode)
   {
   gsm_insisting=1;
   gsm_insist_success=0;
   gsm_insist_starttime=LPC_TIM3->TC;
   gsm_insist_timeout=timeout_us;
   gsm_insist_repeatcount=repeatcount;
   gsm_insist_successmode=successmode;
   gsm_insist_successsubmode=successsubmode;
   gsm_insist_failmode=failmode;
   gsm_insist_failsubmode=failsubmode;
   debug.printf("Insisting:%sHoping for:%s",command,hope);
   if (command!=gsm_insist_command) strcpy(gsm_insist_command,command); // the caller can place the string to the gsm_insist_command itself
   strcpy(gsm_insist_hope,hope);
   gsm.printf("%s",command);
   gsm_wait=timeout_us; 
   }
 


// no insisting during trying or vice versa
int gsm_try(char *command, char *hope, int timeout_us)
   {
   long newtime,lasttime;
   gsm_insist_success=0;
   lasttime=LPC_TIM3->TC;
   debug.printf("Trying:%sHoping for:%s",command,hope);
   strcpy(gsm_insist_hope,hope);
   gsm_insisting=1;
   gsm.printf("%s",command);
   while((gsm_insist_success==0)&&(timeout_us>0))
      {
      newtime=LPC_TIM3->TC;
      timeout_us-=(newtime-lasttime);
      lasttime=newtime;
      }
   gsm_insisting=0;
   return(gsm_insist_success);
   }
   
   
    
void gsm_endcall(void)
   {
   gsm.printf("ATH\r");
   gsm_hangup=1;
   }



void gsm_acceptcall(void)
   {
   gsm_hangup=0;
   gsm.printf("ATA\r");
   }
   
   
   
void gsm_makecall(char *number)
   {
   gsm_hangup=0;
   gsm.printf("ATD%s;\r",number);
   }
 

   
void lzw_init(void);
void lzw_feed(int data, int write);
void lzw_flush(int write);

unsigned char lzwbuffer[128];
int lzwcount,lzwbits,lzwsize=0;


void gsm_sendMMS(char *number)
   {
   int i,j;
   if (gsm_try("AT+CGATT=1\r","OK\r\n",3000000))
      {
      if ((gsm_try("AT+CSTT=\"internet\"\r","OK\r\n",3000000))&& // UPDATE the Access Point Name "internet" with your subscription APN, if necessary
          (gsm_try("AT+CIICR\r","OK\r\n",3000000))&&
          (gsm_try("AT+SAPBR=3,1,\"APN\",\"mms\"\r","OK\r\n",3000000))&& // UPDATE the 'Bearer Context Paremeter' Access Point Name "mms" with your subscription BCP APN, if necessary. I have no idea what that really means. 
          (gsm_try("AT+SAPBR=1,1\r","OK\r\n",3000000)))
         {
         if (gsm_try("AT+CMMSINIT\r","OK\r\n",3000000))
            {
            if ((gsm_try("AT+CMMSCURL=\"mmsc.dnafinland.fi\"\r","OK\r\n",3000000))&& // UPDATE the MultiMedia Service Center URL "mmsc.dnafinland.fi" with your subscription MMSC URL, if necessary
                (gsm_try("AT+CMMSCID=1\r","OK\r\n",3000000))&&
                (gsm_try("AT+CMMSPROTO=\"10.1.1.2\",8080\r","OK\r\n",3000000))&& // UPDATE the MultiMedia Service IP address "10.1.1.2" and port "8080" with those of your subscription, if necessary
                (gsm_try("AT+CMMSEDIT=1\r","OK\r\n",3000000)))
               {
               lzwsize=0; // calculate gif size manually in advance
               lzw_init();
               for(i=0;i<60;i++)
                  {
                  lzw_feed(0x100,0); // clear code
                  for (j=0;j<80;j++) lzw_feed(camera_index_snap[(59-i)*80+(79-j)],0);
                  }
               lzw_feed(0x101,0); //  end of information
               lzw_flush(0);
               sprintf(gsm_insist_command,"AT+CMMSDOWN=\"PIC\",%d,60000,\"thermal.gif\"\r",lzwsize+794);
               if (gsm_try(gsm_insist_command,"CONNECT\r\n",3000000))
                  {
                  gsm_writegif();
                  debug.printf("Data sent\n");
                  gsm_try("","OK\r\n",5000000);
                  sprintf(gsm_insist_command,"AT+CMMSRECP=\"%s\"\r",number);
                  gsm_try(gsm_insist_command,"OK\r\n",3000000);
                  //gsm_try("AT+CMMSRECP=\"harry.tasker@omega.gov\"\r","OK\r\n",3000000); // this is how you set an email recipient for the MMS
                  gsm_try("AT+CMMSSEND\r","OK\r\n",20000000); // this may take long time and the next CMMSTERM had better happen...
                  }
               }
            gsm_try("AT+CMMSTERM\r","OK\r\n",3000000);    
            }
         gsm_try("AT+SAPBR=0,1\r","OK\r\n",3000000);
         }
      gsm_try("AT+CIPSHUT\r","OK\r\n",3000000);
      gsm_try("AT+CGATT=0\r","OK\r\n",3000000);
      }
   }
   

unsigned char gifheader[]={0x47,0x49,0x46,0x38,0x39,0x61, 0x50,0x00, 0x3c,0x00, 0xF7, 0x00, 0x00};
unsigned char gifID[]={0x2C, 0x00,0x00, 0x00,0x00, 0x50,0x00, 0x3c,0x00, 0x00};


void gsm_writegif(void)
   {
   int i,j;
   for(i=0;i<sizeof(gifheader);i++)
      {
      gsm_wait_putc(gifheader[i]);
      }
   for(i=0;i<256;i++) // palette
      {
      gsm_wait_putc((camera_palette[i]&0xf800)>>8);
      gsm_wait_putc((camera_palette[i]&0x07e0)>>3);
      gsm_wait_putc((camera_palette[i]&0x001f)<<3);
      }
   for(i=0;i<sizeof(gifID);i++)
      {
      gsm_wait_putc(gifID[i]);
      }
   gsm_wait_putc(0x08); // LZW minimum code size - data will be 9 bits per pixel. 8 would be easier to encode, but might not be as compatible
   lzw_init();
   for(i=0;i<60;i++)
      {
      lzw_feed(0x100,1); // clear code
      for (j=0;j<80;j++) lzw_feed(camera_index_snap[(59-i)*80+(79-j)],1);
      }
   lzw_feed(0x101,1); //  end of information
   lzw_flush(1);
   gsm_wait_putc(0x00); // end of image data
   gsm_wait_putc(0x3b); // end
   }



void lzw_init(void)
   {
   lzwbuffer[0]=0;
   lzwcount=0;
   lzwbits=0;
   }

void lzw_feed(int data, int write)
   {
   int k;
   for(k=0;k<9;k++)
      {
      lzwbuffer[lzwcount]|=(((data>>k)&0x01)<<lzwbits++);
      if (lzwbits==8)
         {
         lzwcount++;
         lzwbuffer[lzwcount]=0;
         lzwbits=0;
         if (lzwcount==77) lzw_flush(write);
         }
      }
   }

void lzw_flush(int write)
   {
   int i;
   if (lzwbits) lzwcount++;
   if (write)
      {
      gsm_wait_putc(lzwcount);
      for(i=0;i<lzwcount;i++) // lzwcount can be zero
         {
         gsm_wait_putc(lzwbuffer[i]);
         }
      }
   lzwsize+=1+lzwcount;
   lzw_init();
   }


// currently there is only one fixed size image to write
// the data is output slowly to avoid overrun, but later the flow control should be used for this (when it's wired first...)
unsigned char tifheader[]={0x49,0x49, 0x2a,0x00, 0x5e,0x38,0x00,0x00,
                           0x08,0x00, 0x08,0x00, 0x08,0x00, // 8,8,8
                           0xff,0x76,0x01,0x00, 0xe8,0x03,0x00,0x00, // 95999 / 1000
                           0xff,0x76,0x01,0x00, 0xe8,0x03,0x00,0x00}; // 95999 / 1000
unsigned char tiffooter[]={0x0c,0x00,
                           0x00,0x01, 0x04,0x00, 0x01,0x00,0x00,0x00, 0x50,0x00,0x00,0x00, // width 80
                           0x01,0x01, 0x04,0x00, 0x01,0x00,0x00,0x00, 0x3c,0x00,0x00,0x00, // height 60
                           0x02,0x01, 0x03,0x00, 0x03,0x00,0x00,0x00, 0x08,0x00,0x00,0x00, // depth at address 8
                           0x03,0x01, 0x03,0x00, 0x01,0x00,0x00,0x00, 0x01,0x00, 0x00,0x00, // compression 1
                           0x06,0x01, 0x03,0x00, 0x01,0x00,0x00,0x00, 0x02,0x00, 0x00,0x00, // 'photometric' RGB
                           0x11,0x01, 0x04,0x00, 0x01,0x00,0x00,0x00, 0x1e,0x00,0x00,0x00, // stripoffset(s) at 0x0000001e
                           0x15,0x01, 0x03,0x00, 0x01,0x00,0x00,0x00, 0x03,0x00, 0x00,0x00, // samplesperpixel 3 (RGB)
                           0x16,0x01, 0x04,0x00, 0x01,0x00,0x00,0x00, 0x3c,0x00,0x00,0x00, // rowsperstrip 60
                           0x17,0x01, 0x04,0x00, 0x01,0x00,0x00,0x00, 0x40,0x38,0x00,0x00, // stripbytecounts 14400 (80*60*3)
                           0x1a,0x01, 0x05,0x00, 0x01,0x00,0x00,0x00, 0x0e,0x00,0x00,0x00, // xresolution at 0x0000000e
                           0x1b,0x01, 0x05,0x00, 0x01,0x00,0x00,0x00, 0x16,0x00,0x00,0x00, // yresolution at 0x00000016
                           0x28,0x01, 0x03,0x00, 0x01,0x00,0x00,0x00, 0x02,0x00, 0x00,0x00}; // resolutionunit 2 ()
void gsm_writetif(void)
   {
   int i;
   for(i=0;i<sizeof(tifheader);i++)
      {
      gsm_wait_putc(tifheader[i]);
      }
   for(i=0;i<80*60;i++)
      {
      gsm_wait_putc((camera_palette[camera_index_snap[i]]&0xf800)>>8);
      gsm_wait_putc((camera_palette[camera_index_snap[i]]&0x07e0)>>3);
      gsm_wait_putc((camera_palette[camera_index_snap[i]]&0x001f)<<3);
      }
   for(i=0;i<sizeof(tiffooter);i++)
      {
      gsm_wait_putc(tiffooter[i]);
      }
   }
   
   
   
void gsm_wait_putc(char data)
   {
   if (GETIO(GSM_CTS))
      {
      debug.printf("Waiting (CTS=1)\n");
      while(GETIO(GSM_CTS));
      }
   gsm.putc(data);
   }  
   
   
   
// this function is only used to set the phone ready for SIM900D firmware
// update. Call it immediately at program start!!! It does not return,
// so the system needs to be power cycled (and the ARM reflashed)
// It is preferable to update the SIM900D before it is soldered to the board,
// so there is no need to use this at all
void gsm_flash(void)
   {

   system_configurepin(GSM_PWRKEY, 0, 0, 1, 1);    // p73, P2.2, 0 (GPIO), pull-up, output, 1
   while(1)
      {
      if (system_debugpin()) SETIO(GSM_PWRKEY);
      else CLRIO(GSM_PWRKEY);
      wait_ms(100);
      }
      
   // the earlier try, but the module does not seem to accept update from the main serial port - only debug serial port    
   CLRIO(GSM_PWRKEY);
   debug.baud(460800);
   gsm.baud(460800);

   while(1)
      {
      if (gsm.readable()) debug.putc(gsm.getc());
      if (debug.readable()) gsm.putc(debug.getc());
      }
   }

     